#ifdef VXWORKS
#  include <vxWorks.h>
#  include <stdioLib.h>
#  include <logLib.h>
#else 
#  include <stdio.h>
#  include <portLib.h>
#  define logMsg printf
#endif
#include <math.h>
#include <stdlib.h>
#include "h2mathLib.h"
#include "genPosStruct.h"
#include "contlaw/contLaw.h"

static void errorsGet (GENPOS_CART_CMD_PARAM *cmdId,
		       GENPOS_CART_REF *refId,
		       GENPOS_CART_STATE *posId,
		       double period,
		       double *pdLongit, /* Ou` mettre l'ecart longitudinal */
		       double *pdTransv,  /* Ou` mettre l'ecart transversal */
		       double *pSdLongit, /* Ou` mettre somme ecarts longit. */
		       double *pSdTransv); /* Ou` mettre somme ecarts transv. */

static void errorsConfigGet(GENPOS_CART_CMD_PARAM *cmdId,
			    GENPOS_CART_REF *refId,
			    GENPOS_CART_STATE *posId,
			    double period,
			    double *pdLongit, /* Ecart longitudinal */
			    double *pdTransv,  /* Ecart transversal */
			    double *pdTheta,   /* Ecart angulaire */
			    double *pSdLongit, /* Somme ecarts longit. */
			    double *pSdTransv, /* Somme ecarts transv. */
			    double *pSdTheta); /* Somme ecarts ang. */
static double sinc(double x);

/*========================= SERVO CONTROL LAWS  =========================*/


/******************************************************************************
*
*  contLawCartPositionControl  -  
*
*  Description:
*  This function computes a control law for stabilizing a cart-like robot
*  along a trajectory. The control law consists of a PID controller applied
*  to a point in front of the wheel axis of the robot.
*
*  Return: importantDriftFlag
*/

int contLawCartPositionControl (GENPOS_CART_CMD_PARAM *cmdId,
				   GENPOS_CART_REF *refId,
				   GENPOS_CART_STATE *posId,
				   double period, 
				   double prevV, double prevW,
				   double *vRef, double *wRef)

{
  double dLongit;               /* Erreur de deplacement longitudinal */
  double dTransv;               /* Erreur de deplacement transversal */
  double SdLongit;              /* Somme des erreurs longitudinales */
  double SdTransv;              /* Somme des erreurs transverses */
  double sense;                 /* Sens de deplacement */
  double v;                     /* Vitesse lineaire de consigne */
  double w;                     /* Vitesse angulaire de consigne */
  double dv;                    /* Variation de vitesse lineaire */
  double dw;                    /* Variation de vitesse angulaire */
  int importantDriftFlag;

  /* Obtenir le sens de deplacement */
  sense = refId->backFlag == GENPOS_BACKWARD_MOTION ? -1 : 1;

  /* Obtenir les ecarts de position et leurs integrales */
  errorsGet (cmdId, refId, posId, period, &dLongit, &dTransv, &SdLongit, &SdTransv);

  /* Verifier que le robot suit bien la consigne */
  if (fabs(dLongit) > cmdId->longitMaxGap ||
      fabs(dTransv) > cmdId->transvMaxGap) 
    importantDriftFlag = TRUE;
  else
    importantDriftFlag = FALSE;
  
  /* Commande de vitesse lineaire a donner au robot */
  v = refId->v + (cmdId->KpLongit*dLongit*refId->linAccelMax)/refId->vmax +
    cmdId->KiLongit * SdLongit;

  /* Limiter l'acceleration lineaire */
  if (fabs (dv = v - prevV) > refId->linAccelMax * period)
    v = prevV +  SIGN (dv) * refId->linAccelMax * period;
  
  /* Limiter la vitesse lineaire */
  if (fabs (v) > refId->vmax)
    v = SIGN (v) * refId->vmax;

  /* Memorisation de la vitesse lineaire */
  prevV = v;
  
  /* Commande de la vitesse angulaire */
  w = refId->w +
    sense *(cmdId->KpTransv*dTransv*refId->angAccelMax/refId->wmax +
	    cmdId->KiTransv*SdTransv);
  
  /* Limiter l'acceleration angulaire */
  if (fabs (dw = w - prevW) > refId->angAccelMax * period)
    w = prevW + SIGN (dw) * refId->angAccelMax * period;

  /* Limiter la vitesse angulaire */
  if (fabs (w) >  refId->wmax)
    w = SIGN (w) * refId->wmax;

  /* Memorisation de la vitesse angulaire */
  prevW = w;

  /* Imposer au robot les vitesses lineaire et angulaire de consigne */
  *vRef = v;
  *wRef = w;
  return importantDriftFlag;
}


/******************************************************************************
*
*  contLawCartConfigControl  -  
*
*  Description: This function computes a control law for stabilizing a
*  cart-like robot along a trajectory. The control law takes into
*  account the three configuration variables of the robot.
*
*  Return: importantDriftFlag
*/

STATUS contLawCartConfigControl (GENPOS_CART_CMD_PARAM *cmdId,
				    GENPOS_CART_REF *refId,
				    GENPOS_CART_STATE *posId, 
				    double period,
				    double prevV, double prevW,
				    double *vRef, double *wRef)

{
  double dLongit;               /* Erreur de deplacement longitudinal */
  double dTransv;               /* Erreur de deplacement transversal */
  double dTheta;                /* Erreur angulaire */
  double SdLongit;              /* Somme des erreurs longitudinales */
  double SdTransv;              /* Somme des erreurs transverses */
  double SdTheta;               /* Somme des erreurs angulaires */
  double v, v0 = refId->v;      /* Vitesse lineaire de consigne */
  double w;                     /* Vitesse angulaire de consigne */
  double dv;                    /* Variation de vitesse lineaire */
  double dw;                    /* Variation de vitesse angulaire */

  int importantDriftFlag;

  /* Obtenir les ecarts de position et leurs integrales */
  errorsConfigGet(cmdId, refId, posId, period,
		  &dLongit, &dTransv, &dTheta, 
		  &SdLongit, &SdTransv, &SdTheta);

  /* Verifier que le robot suit bien la consigne */
  if (fabs(dLongit) > cmdId->longitMaxGap ||
      fabs(dTransv) > cmdId->transvMaxGap ||
      fabs(dTheta)  > cmdId->angMaxGap) 
    importantDriftFlag = TRUE;
  else
    importantDriftFlag = FALSE;
					  
  /* Commande de vitesse lineaire a donner au robot */
  v = v0*cos(dTheta) + 
    cmdId->KpLongit*dLongit +cmdId->KiLongit*SdLongit;
  /* Limiter l'acceleration lineaire */
  if (fabs (dv = v - prevV) > refId->linAccelMax * period)
    v = prevV +  SIGN (dv) * refId->linAccelMax * period;
  
  /* Limiter la vitesse lineaire */
  if (fabs (v) > refId->vmax)
    v = SIGN (v) * refId->vmax;

  /* Memorisation de la vitesse lineaire */
  prevV = v;
  
  /* Commande de la vitesse angulaire */
  w = refId->w +
    cmdId->KpAng*dTheta+cmdId->KiAng*SdTheta+
    v0 * sinc(dTheta)*(cmdId->KpTransv*dTransv+
		       cmdId->KiTransv*SdTransv);

  /* Limiter l'acceleration angulaire */
  if (fabs (dw = w - prevW)  > refId->angAccelMax * period)
    w = prevW + SIGN (dw) * refId->angAccelMax * period;

  /* Limiter la vitesse angulaire */
  if (fabs (w) >  refId->wmax)
    w = SIGN (w) * refId->wmax;

  /* Memorisation de la vitesse angulaire */
  prevW = w;

  /* Imposer au robot les vitesses lineaire et angulaire de consigne */
  *vRef = v;
  *wRef = w;
  return importantDriftFlag;
}


/******************************************************************************
*
*  contralLawCartSpeedControl  -  
*
*  Description:
*  This function controls the speed of a cart-like robot. The input speed
*  is corrected to take into account maximal accelerations and velocities.
*
*  Return: importantDriftFlag
*/

int contralLawCartSpeedControl(GENPOS_CART_SPEED *refId,
			       double period, 
			       double prevV, double prevW,
			       GENPOS_CART_SPEED *refOutId)
{
#if 0
  double v;                     /* Vitesse lineaire de consigne */
  double w;                     /* Vitesse angulaire de consigne */
  double dv;                    /* Variation de vitesse lineaire */
  double dw;                    /* Variation de vitesse angulaire */
#endif
  int importantDriftFlag=FALSE;

  *refOutId = *refId;

  /* XXXX va disparaitre car redondant avec locoVelocityControl */
  /* NON, pas forcement redondant car peut permettre de limiter 
     les vitesses en restant sur la trajectoire (ie, en conservant 
     le rapport v/w). Regarder ce qui a ete fait dans lloco de lama. */
#if 0
  /* Commande de vitesse lineaire a donner au robot */
  v = refId->v;

  /* Limiter l'acceleration lineaire */
  if (fabs (((dv = v - prevV) / period)) > AMAX_LIN_ROBOT)
    v = prevV +  SIGN (dv) * AMAX_LIN_ROBOT * period;
  
  /* Limiter la vitesse lineaire */
  if (fabs (v) > VMAX_ROBOT)
    v = SIGN (v) * VMAX_ROBOT;

  /* Memorisation de la vitesse lineaire */
  prevV = v;
  
  /* Commande de la vitesse angulaire */
  w = refId->w;

  /* Limiter l'acceleration angulaire */
  if (fabs ((dw = w - prevW) / period) > AMAX_ANG_ROBOT)
    w = prevW + SIGN (dw) * AMAX_ANG_ROBOT * period;

  /* Limiter la vitesse angulaire */
  if (fabs (w) >  WMAX_ROBOT)
    w = SIGN (w) * WMAX_ROBOT;

  /* Memorisation de la vitesse angulaire */
  prevW = w;

  refOutId->v = v;
  refOutId->w = w;
#endif

  prevV = refId->v;
  prevW = refId->w;

  return importantDriftFlag;
}




/*============================= LOCAL FUNCTIONS  =========================*/


/*****************************************************************************
*
*  errorsGet  -  
*
*  Description: 
*  This function is called by contLawCartPositionControl. It computes 
*  the gaps between the reference position and the real position of 
*  the control point.
*
* Return: Neant
*/

static void errorsGet (GENPOS_CART_CMD_PARAM *cmdId,
		       GENPOS_CART_REF *refId,
		       GENPOS_CART_STATE *posId,
		       double period,
		       double *pdLongit, /* Ou` mettre l'ecart longitudinal */
		       double *pdTransv,  /* Ou` mettre l'ecart transversal */
		       double *pSdLongit, /* Ou` mettre somme ecarts longit. */
		       double *pSdTransv) /* Ou` mettre somme ecarts transv. */

{
  double dxA, dyA;              /* Erreurs position par rapport rep atelier */
  double dLongit;               /* Erreur de deplacement longitudinal */
  double dTransv;               /* Erreur de deplacement transversal */
  static double SdLongit = 0.;   /* Somme des erreurs longitudinales */
  static double SdTransv = 0.;   /* Somme des erreurs transverses */
  double c, s;


  /* Donne'es positionnelles sur les quelles on s'asservit */
  c = cos(posId->theta);
  s = sin(posId->theta);

  /* Obtenir les erreurs de position par rapport au rep atelier */
  dxA = refId->x - posId->xRef;
  dyA = refId->y - posId->yRef;
  
  /* Obtenir les erreurs de deplacement transversal et longitudinal */
  *pdLongit = dLongit =   dxA * c + dyA * s;
  *pdTransv = dTransv = - dxA * s + dyA * c;

  /* Actualiser l'integrale des erreurs longitudinales */
  if (fabs (dLongit) < cmdId->longitErrorMax) {

    /* Actualiser la somme des erreurs */
    SdLongit += dLongit*period;
    
    /* La limiter */
    if (fabs (SdLongit) > cmdId->longitMaxErrSum)
      *pSdLongit = SdLongit = SIGN(SdLongit) * cmdId->longitMaxErrSum;
    else 
      *pSdLongit = SdLongit;
  }
  else 
    *pSdLongit = SdLongit = 0.;

  /* Actualiser l'integrale des erreurs transversales */
  if (fabs (dTransv) < cmdId->transvErrorMax) {
    
    /* Actualiser la somme des erreurs */
    SdTransv += dTransv*period;
    
    /* La limiter */
    if (fabs (SdTransv) > cmdId->transvMaxErrSum)
      *pSdTransv = SdTransv = SIGN(SdTransv) * cmdId->transvMaxErrSum;
    else
      *pSdTransv = SdTransv;
  }
  else 
    *pSdTransv = SdTransv = 0.;
}


/*****************************************************************************
*
*  errorsConfigGet  -  
*
*  Description:
*  This function is called by contLawCartConfigControl. It computes 
*  the gaps between the reference configuration and the real configuration 
*  of a cart-like robot.
*
*  Return: Neant
*/

static void errorsConfigGet(GENPOS_CART_CMD_PARAM *cmdId,
			    GENPOS_CART_REF *refId,
			    GENPOS_CART_STATE *posId,
			    double period,
			    double *pdLongit, /* Ecart longitudinal */
			    double *pdTransv,  /* Ecart transversal */
			    double *pdTheta,   /* Ecart angulaire */
			    double *pSdLongit, /* Somme ecarts longit. */
			    double *pSdTransv, /* Somme ecarts transv. */
			    double *pSdTheta) /* Somme ecarts ang. */

{
  double dxA, dyA;              /* Erreurs position par rapport rep atelier */
  double dLongit;               /* Erreur de deplacement longitudinal */
  double dTransv;               /* Erreur de deplacement transversal */
  double dTheta=0;                /* Erreur de deplacement angulaire */
  static double SdLongit = 0.;   /* Somme des erreurs longitudinales */
  static double SdTransv = 0.;   /* Somme des erreurs transverses */
  static double SdTheta = 0.;      /* Somme des erreurs angulaires */
  double c, s;

  /* Donne'es positionnelles sur lesquelles on s'asservi */
  c = cos(posId->theta);
  s = sin(posId->theta);

  /* Obtenir les erreurs de position par rapport au rep atelier */
  /* point de controle */
  dxA = refId->x - posId->xRob;
  dyA = refId->y - posId->yRob;
  
  /* angles */
  *pdTheta = angleLimit(refId->theta - posId->theta);

  /* Obtenir les erreurs de deplacement transversal et longitudinal */
  *pdLongit = dLongit =   dxA * c + dyA * s;
  *pdTransv = dTransv = - dxA * s + dyA * c;

  /* Actualiser l'integrale des erreurs longitudinales */
  if (fabs (dLongit) < cmdId->longitErrorMax) {

    /* Actualiser la somme des erreurs */
    SdLongit += dLongit*period;
    
    /* La limiter */
    if (fabs (SdLongit) > cmdId->longitMaxErrSum)
      *pSdLongit = SdLongit = cmdId->longitMaxErrSum;
    else 
      *pSdLongit = SdLongit;
  }
  else 
    *pSdLongit = SdLongit = 0.;

  /* Actualiser l'integrale des erreurs transversales */
  if (fabs (dTransv) < cmdId->transvErrorMax) {
    
    /* Actualiser la somme des erreurs */
    SdTransv += dTransv*period;
    
    /* La limiter */
    if (fabs (SdTransv) > cmdId->transvMaxErrSum)
      *pSdTransv = SdTransv = cmdId->transvMaxErrSum;
    else
      *pSdTransv = SdTransv;
  }
  else 
    *pSdTransv = SdTransv = 0.;

  /* Actualiser l'integrale des erreurs angulaires */
  if (fabs (dTheta) < cmdId->angErrorMax) {
    
    /* Actualiser la somme des erreurs */
    SdTheta += dTheta*period;
    
    /* La limiter */
    if (fabs (SdTheta) > cmdId->angMaxErrSum)
      *pSdTheta = SdTheta = cmdId->angMaxErrSum;
    else
      *pSdTheta = SdTheta;
  }
  else 
    *pSdTheta = SdTheta = 0.;
}




/**
 **  sinc  
 **
 **/
static double sinc(double x)
{
  double y=1.0;

  if (fabs(x)>1e-8)
    {
      y = sin(x)/x;
    }
  return y;
}
