#ifdef VXWORKS
#  include <vxWorks.h>
#  include <stdioLib.h>
#  include <logLib.h>
#else
#  include <stdio.h>
#  include <portLib.h>
#  define logMsg printf
#endif
#include <math.h>
#include <stdlib.h>
#include "h2mathLib.h"
#include "genPosStruct.h"
#include "contlaw/contLaw.h"

static void errorsConfigRobotGet(GENPOS_CART_CMD_PARAM *cmdId,
				 GENPOS_CART_REF *refId,
				 GENPOS_CART_CONFIG *posId, 
				 double period,
				 double *pdLongit, /* Ecart longitudinal */
				 double *pdTransv,  /* Ecart transversal */
				 double *pdTheta,   /* Ecart angulaire */
				 double *pSdLongit, /* Somme ecarts longit. */
				 double *pSdTransv, /* Somme ecarts transv. */
				 double *pSdTheta); /* Somme ecarts ang. */

static void 
errorVirtualRobotGet(GENPOS_CART_CMD_PARAM *cmdId,
		     GENPOS_TRCART_REF *refId,
		     GENPOS_TRCART_CONFIG *posId,
		     double period,
		     GENPOS_CART_STATE * virtualRobotId,
		     GENPOS_CART_REF * virtualRefId,
		     double *pdLongit, /* Ecart longitudinal */
		     double *pdTransv,  /* Ecart transversal */
		     double *pdTheta2,  /* Ecart ang robot virtuel */
		     double *pSdLongit, /* Somme ecarts longit. */
		     double *pSdTransv, /* Somme ecarts transv. */
		     double *pSdTheta2); /* Somme ecarts ang rob virt*/


static void 
virtualRobotPositionGetOffAxle(GENPOS_TRAILER_CONNECTION *kinParamPt,
			       GENPOS_TRCART_REF *refRobotId,
			       GENPOS_TRCART_CONFIG * posId,
			       GENPOS_CART_STATE * virtualRobotId,
			       GENPOS_CART_REF * virtualRefId);

static void 
virtualRobotPositionGetCenteredAxle(GENPOS_TRAILER_CONNECTION *kinParamPt,
				    GENPOS_TRCART_REF *refRobotId,
				    GENPOS_TRCART_CONFIG * posId,
				    GENPOS_CART_STATE * virtualRobotId,
				    GENPOS_CART_REF * virtualRefId);

/**
 **  sinc  
 **
 **/
static double sinc(double x)
{
  double y=1.0;

  if (fabs(x)>1e-8)
    {
      y = sin(x)/x;
    }
  return y;
}



/*
 *  contLawCartTrailerConfigControl --
 *
 *  control law for a cart towing a trailer.
 *  Two types of trailer connections are considered:
 *    GENPOS_TRAILER_OFF_AXLE when the trailer is connected behind
 *    the wheel axis of the robot and 
 *    GENPOS_TRAILER_CENTERED_AXLE when the robot is connected above
 *    the wheel axis of the robot.
 *
 *  When the robot moves forward, the trailer is not stabilized and 
 *  the same control law as in contLawCartConfigControl is applied.
 *  When the robot goes backward, the above control law is applied 
 *  to a virtual robot.
 */

int contLawCartTrailerConfigControl(GENPOS_TRAILER_CONNECTION *kinParamPt,
				    GENPOS_CART_CMD_PARAM *cmdId,
				    GENPOS_TRCART_REF *refId,
				    GENPOS_TRCART_CONFIG *posId,
				    double period, 
				    GENPOS_TRAILER_TYPE trailerType,
				    double prevV, double prevW, 
				    double *vRefPt, double *wRefPt)
{
  int importantDriftFlag;

  if (refId->robot.backFlag == GENPOS_FORWARD_MOTION)
    importantDriftFlag = 
      contLawTrailerForwardControl(cmdId, &(refId->robot), &(posId->robot),
				   period, prevV, prevW, vRefPt, wRefPt);
  else
    if (trailerType == GENPOS_TRAILER_OFF_AXLE)
      importantDriftFlag = 
	contLawOffAxleTrBackwardControl(kinParamPt,
					cmdId, refId, posId,
					period, 
					prevV, prevW,
					vRefPt, wRefPt);
    else 
      importantDriftFlag=
	contLawCentAxleTrBackwardControl(kinParamPt,
					 cmdId, refId, posId,
					 period, 
					 prevV, prevW,
					 vRefPt, wRefPt);
  return importantDriftFlag;
}






/******************************************************************************
*
*  contLawTrailerForwardControl  -  
*
*  Description:
*  This function controls a cart towing a trailer going forward. Only
*  the cart is stabilized. The trailer is intrinsically stable. The 
*  control law is the same as in contLawCartConfigControl.
*
*  Return: importantDriftFlag
*/

STATUS contLawTrailerForwardControl (GENPOS_CART_CMD_PARAM *cmdId,
				     GENPOS_CART_REF *refId,
				     GENPOS_CART_CONFIG *posId, 
				     double period,
				     double prevV, double prevW,
				     double *vRef, double *wRef)

{
  double dLongit;               /* Erreur de deplacement longitudinal */
  double dTransv;               /* Erreur de deplacement transversal */
  double dTheta;                /* Erreur angulaire */
  double SdLongit;              /* Somme des erreurs longitudinales */
  double SdTransv;              /* Somme des erreurs transverses */
  double SdTheta;               /* Somme des erreurs angulaires */
  double v, v0 = refId->v;      /* Vitesse lineaire de consigne */
  double w;                     /* Vitesse angulaire de consigne */
  double dv;                    /* Variation de vitesse lineaire */
  double dw;                    /* Variation de vitesse angulaire */

  int importantDriftFlag;

  /* Obtenir les ecarts de position et leurs integrales */
  errorsConfigRobotGet(cmdId, refId, posId, period, 
		       &dLongit, &dTransv, &dTheta, 
		       &SdLongit, &SdTransv, &SdTheta);

  /* Verifier que le robot suit bien la consigne */
  if (fabs(dLongit) > cmdId->longitMaxGap ||
      fabs(dTransv) > cmdId->transvMaxGap ||
      fabs(dTheta)  > cmdId->angMaxGap) 
    importantDriftFlag = TRUE;
  else
    importantDriftFlag = FALSE;
					  
  /* Commande de vitesse lineaire a donner au robot */
  v = v0*cos(dTheta) + 
    cmdId->KpLongit*dLongit +cmdId->KiLongit*SdLongit;
  /* Limiter l'acceleration lineaire */
  if (fabs (dv = v - prevV) > refId->linAccelMax * period)
    v = prevV +  SIGN (dv) * refId->linAccelMax * period;
  
  /* Limiter la vitesse lineaire */
  if (fabs (v) > refId->vmax)
    v = SIGN (v) * refId->vmax;

  /* Commande de la vitesse angulaire */
  w = refId->w +
    cmdId->KpAng*dTheta+cmdId->KiAng*SdTheta+
    v0 * sinc(dTheta)*(cmdId->KpTransv*dTransv+
		       cmdId->KiTransv*SdTransv);

  /* Limiter l'acceleration angulaire */
  if (fabs (dw = w - prevW) > refId->angAccelMax * period)
    w = prevW + SIGN (dw) * refId->angAccelMax * period;

  /* Limiter la vitesse angulaire */
  if (fabs (w) >  refId->wmax)
    w = SIGN (w) * refId->wmax;

  /* Imposer au robot les vitesses lineaire et angulaire de consigne */
  *vRef = v;
  *wRef = w;
  return importantDriftFlag;
}

/******************************************************************************
*
*  contLawOffAxleTrBackwardControl  -  
*
*  Description:
*  This function controls a cart pushing a trailer backward. The trailer
*  is hitched behind the wheel axis of the cart. 
*  The function virtually applies the control law implemented in 
*  contLawCartConfigControl to the trailer.
*
*  Return: importantDriftFlag
*/

int contLawOffAxleTrBackwardControl(GENPOS_TRAILER_CONNECTION *kinParamPt,
				    GENPOS_CART_CMD_PARAM *cmdId,
				    GENPOS_TRCART_REF *refId,
				    GENPOS_TRCART_CONFIG *posId,
				    double period,
				    double prevV, double prevW,
				    double *vRef, double *wRef)
{
  GENPOS_CART_REF *refRobotId = &(refId->robot);
  double dLongit;               /* Erreur de deplacement longitudinal */
  double dTransv;               /* Erreur de deplacement transversal */
  double dTheta;                /* Erreur angulaire robot virtuel*/
  double SdLongit;              /* Somme des erreurs longitudinales */
  double SdTransv;              /* Somme des erreurs transverses */
  double SdTheta;               /* Somme des erreurs ang robot virtuel */
  double v;                     /* Vitesse lineaire de consigne */
  double w;                     /* Vitesse angulaire de consigne */
  double dv;                    /* Variation de vitesse lineaire */
  double dw;                    /* Variation de vitesse angulaire */
  double l1 = kinParamPt->l1;
  double l2 = kinParamPt->l2; 
  /* variables intermediaires de calcul */
  double phi = posId->phi; /* angle reel */
  double v2, v2Ref, w2, w2Ref;  /* vitesses du robot virtuel */
  GENPOS_CART_STATE virtualRobot;
  GENPOS_CART_REF virtualRef;
  int importantDriftFlag;

  virtualRobotPositionGetOffAxle(kinParamPt, refId, posId,
				 &virtualRobot, &virtualRef);

  /* Obtenir les ecarts de position et leurs integrales */
  errorVirtualRobotGet(cmdId, refId, posId, period, &virtualRobot,
		       &virtualRef, &dLongit, &dTransv, &dTheta, 
		       &SdLongit, &SdTransv, &SdTheta);
  
  /* Verifier que le robot suit bien la consigne */
  if (fabs(dLongit) > cmdId->longitMaxGap ||
      fabs(dTransv) > cmdId->transvMaxGap ||
      fabs(dTheta)  > cmdId->angMaxGap)
    importantDriftFlag = TRUE;
  else
    importantDriftFlag = FALSE;
  
  v2Ref = virtualRef.v;
  w2Ref = virtualRef.w;

  v2 = v2Ref*cos(dTheta) +
    cmdId->KpLongit*dLongit+cmdId->KiLongit*SdLongit;

  w2 = w2Ref + 
    cmdId->KpAng*dTheta + cmdId->KiAng*SdTheta +
    v2Ref*sinc(dTheta)*(cmdId->KpTransv*dTransv+cmdId->KiTransv*SdTransv);

  v = -v2*cos(phi)-l2*w2*sin(phi);
  w = sin(phi)/l1*v2 - l2*cos(phi)*w2/l1;

  /* Limiter l'acceleration lineaire */
  if (fabs (dv = v - prevV) > refRobotId->linAccelMax * period)
    v = prevV +  SIGN (dv) * refRobotId->linAccelMax * period;
  
  /* Limiter la vitesse lineaire */
  if (fabs (v) > refRobotId->vmax)
    v = SIGN (v) * refRobotId->vmax;
  
  /* Memorisation de la vitesse lineaire */
  prevV = v;
  
  /* Limiter l'acceleration angulaire */
  if (fabs (dw = w - prevW) > refRobotId->angAccelMax * period)
    w = prevW + SIGN (dw) * refRobotId->angAccelMax * period;
  
  /* Limiter la vitesse angulaire */
  if (fabs (w) >  refRobotId->wmax)
    w = SIGN (w) * refRobotId->wmax;
  
  /* Memorisation de la vitesse angulaire */
  prevW = w;
  
  /* Imposer au robot les vitesses lineaire et angulaire de consigne */
  *vRef = v;
  *wRef = w;
  return importantDriftFlag;
}

/* 
 *   contLawCentAxleTrBackwardControl --
 *
 *  Description:
 *  This function controls a cart pushing a trailer backward. The trailer
 *  is hitched above the wheel axis of the cart. 
 *  The function virtually applies the control law implemented in 
 *  contLawCartConfigControl to a virtual robot, symmetric of the real
 *  robot wrt the wheel axis of the trailer.
 *
 *  Return: importantDriftFlag
 */

int contLawCentAxleTrBackwardControl(GENPOS_TRAILER_CONNECTION *kinParamPt,
				     GENPOS_CART_CMD_PARAM *cmdId,
				     GENPOS_TRCART_REF *refId,
				     GENPOS_TRCART_CONFIG *posId,
				     double period,
				     double prevV, double prevW,
				     double *vRef, double *wRef)
{
  GENPOS_CART_REF *refRobotId = &(refId->robot);
  double dLongit;               /* Erreur de deplacement longitudinal */
  double dTransv;               /* Erreur de deplacement transversal */
  double dTheta;                /* Erreur angulaire robot virtuel*/
  double SdLongit;              /* Somme des erreurs longitudinales */
  double SdTransv;              /* Somme des erreurs transverses */
  double SdTheta;               /* Somme des erreurs ang robot virtuel */
  double v;                     /* Vitesse lineaire de consigne */
  double w;                     /* Vitesse angulaire de consigne */
  double dv;                    /* Variation de vitesse lineaire */
  double dw;                    /* Variation de vitesse angulaire */
  double l2 = kinParamPt->l2; 
  /* variables intermediaires de calcul */
  double v0 = refRobotId->v;          /* vitesse de consigne */
  double phiRef= refId->phi;     /* angle de consigne */
  double phi = posId->phi; /* angle reel */
  GENPOS_CART_STATE virtualRobot;
  GENPOS_CART_REF virtualRef;
  int importantDriftFlag;


  virtualRobotPositionGetCenteredAxle(kinParamPt, refId, posId,
				      &virtualRobot, &virtualRef);
  /* Obtenir les ecarts de position et leurs integrales */
  errorVirtualRobotGet(cmdId, refId, posId, period, &virtualRobot,
		       &virtualRef, &dLongit, &dTransv, &dTheta, 
		       &SdLongit, &SdTransv, &SdTheta);
  
  /* Verifier que le robot suit bien la consigne */
  if (fabs(dLongit) > cmdId->longitMaxGap ||
      fabs(dTransv) > cmdId->transvMaxGap ||
      fabs(dTheta)  > cmdId->angMaxGap)
    importantDriftFlag = TRUE;
  else
    importantDriftFlag = FALSE;
  
  /* Commande vitesse lineaire du robot reel */
  v = v0*(cos(dTheta)+cmdId->KpLongit*dLongit+
	  cmdId->KiLongit*SdLongit); 
  
  /* Commande vitesse angulaire robot reel */
  w = refRobotId->w+v0*(cmdId->KpAng*dTheta+ 
			cmdId->KiAng*SdTheta+
			sinc(dTheta)*(cmdId->KpTransv*dTransv+
				      cmdId->KiTransv*SdTransv))+
    2/l2*(v0*sin(phiRef)-v*sin(phi));

  /* Limiter l'acceleration lineaire */
  if (fabs (dv = v - prevV) > refRobotId->linAccelMax * period)
    v = prevV +  SIGN (dv) * refRobotId->linAccelMax * period;
  
  /* Limiter la vitesse lineaire */
  if (fabs (v) > refRobotId->vmax)
    v = SIGN (v) * refRobotId->vmax;
  
  /* Memorisation de la vitesse lineaire */
  prevV = v;
  
  /* Limiter l'acceleration angulaire */
  if (fabs (dw = w - prevW) > refRobotId->angAccelMax * period)
    w = prevW + SIGN (dw) * refRobotId->angAccelMax * period;
  
  /* Limiter la vitesse angulaire */
  if (fabs (w) >  refRobotId->wmax)
    w = SIGN (w) * refRobotId->wmax;
  
  /* Memorisation de la vitesse angulaire */
  prevW = w;
  
  /* Imposer au robot les vitesses lineaire et angulaire de consigne */
  *vRef = v;
  *wRef = w;
  return importantDriftFlag;
}

/*****************************************************************************
*
*  errorVirtualRobotGet  -  
*
*  Description:
*
*  This function is called by contLawOffAxleTrBackwardControl and
*  contLawCentAxleTrBackwardControl. It computes the gaps between the
*  real and reference configuration of a virtual robot associated to *
*  a cart towing a trailer. If the trailer is hitched behind the wheel
*  axis of the cart, the virtual robot is the trailer. If the trailer
*  is hitched above the wheel axis of the cart, the virtual robot is
*  the symmetric of the cart wrt the wheel axis of the trailer.
*
*  Return: Neant
*/

static void 
errorVirtualRobotGet(GENPOS_CART_CMD_PARAM *cmdId,
		     GENPOS_TRCART_REF *refId,
		     GENPOS_TRCART_CONFIG *posId,
		     double period,
		     GENPOS_CART_STATE * virtualRobotId,
		     GENPOS_CART_REF * virtualRefId,
		     double *pdLongit, /* Ecart longitudinal */
		     double *pdTransv,  /* Ecart transversal */
		     double *pdTheta2,  /* Ecart ang robot virtuel */
		     double *pSdLongit, /* Somme ecarts longit. */
		     double *pSdTransv, /* Somme ecarts transv. */
		     double *pSdTheta2) /* Somme ecarts ang rob virt*/

{
  double dxA, dyA;              /* Erreurs position par rapport rep atelier */
  double dLongit;               /* Erreur de deplacement longitudinal */
  double dTransv;               /* Erreur de deplacement transversal */
  double dTheta=0;                /* Erreur angulaire vrai robot*/
  double dTheta2;               /* Erreur angulaire robot virtuel*/
  static double SdLongit = 0.;   /* Somme des erreurs longitudinales */
  static double SdTransv = 0.;   /* Somme des erreurs transverses */
  static double SdTheta2 = 0.;    /* Somme des erreurs angulaires */
  double theta;                  /* Position du robot reel */

  double x2, y2, theta2;          /* position du robot virtuel */
  double thetaRef, phiRef;        /*consigne reelle*/
  double x2Ref, y2Ref, theta2Ref; /*consigne robot virtuel */
  double c, s, c2, s2;

  /* Position du robot reel */
  theta = posId->robot.theta;

  /* Consigne du robot reel */
  thetaRef = refId->robot.theta;
  phiRef = refId->phi;

  /*Position du robot virtuel */
  x2 = virtualRobotId->xRob;
  y2 = virtualRobotId->yRob;
  theta2 = virtualRobotId->theta;

  /*Consigne du robot virtuel */
  x2Ref = virtualRefId->x;
  y2Ref = virtualRefId->y;
  theta2Ref = virtualRefId->theta;
  
  /*Erreur angulaire vrai robot */
  dTheta = angleLimit(thetaRef - theta);

  /* sin cos */
  c = cos(theta2);
  s = sin(theta2);

  c2 = cos(phiRef + dTheta);
  s2 = sin(phiRef + dTheta);

  /* Ecart angulaire robot virtuel */
  dTheta2 = angleLimit(theta2Ref-theta2);
  *pdTheta2 = dTheta2;

  /* Obtenir les erreurs de position par rapport au rep atelier */
  /* point de controle */
  dxA = x2Ref - x2;
  dyA = y2Ref - y2;
  
  /* Obtenir les erreurs de deplacement transversal et longitudinal */
  *pdLongit = dLongit =   dxA * c + dyA * s;
  *pdTransv = dTransv = - dxA * s + dyA * c;

  /* Actualiser l'integrale des erreurs longitudinales */
  if (fabs (dLongit) < cmdId->longitErrorMax) {

    /* Actualiser la somme des erreurs */
    SdLongit += dLongit*period;
    
    /* La limiter */
    if (fabs (SdLongit) > cmdId->longitMaxErrSum)
      *pSdLongit = SdLongit = cmdId->longitMaxErrSum;
    else 
      *pSdLongit = SdLongit;
  }
  else 
    *pSdLongit = SdLongit = 0.;

  /* Actualiser l'integrale des erreurs transversales */
  if (fabs (dTransv) < cmdId->transvErrorMax) {
    
    /* Actualiser la somme des erreurs */
    SdTransv += dTransv*period;
    
    /* La limiter */
    if (fabs (SdTransv) > cmdId->transvMaxErrSum)
      *pSdTransv = SdTransv = cmdId->transvMaxErrSum;
    else
      *pSdTransv = SdTransv;
  }
  else 
    *pSdTransv = SdTransv = 0.;

  /* Actualiser l'integrale des erreurs angulaires */
  if (fabs (dTheta2) < cmdId->angErrorMax) {
    
    /* Actualiser la somme des erreurs */
    SdTheta2 += dTheta2*period;
    
    /* La limiter */
    if (fabs (SdTheta2) > cmdId->angMaxErrSum)
      *pSdTheta2 = SdTheta2 = cmdId->angMaxErrSum;
    else
      *pSdTheta2 = SdTheta2;
  }
  else 
    *pSdTheta2 = SdTheta2 = 0.;
}

/*****************************************************************************
*
*  errorsConfigGet  -  Obtenir les ecarts de position du robot
*
*  Description:

*  This function is called by contLawTrailerForwardControl. It
*  computes the gaps between the real and reference configuration of a
*  cart towing a trailer. Only the variable configuration of the cart
*  are consisdered.
*

*  Return: Neant
*/

static void errorsConfigRobotGet(GENPOS_CART_CMD_PARAM *cmdId,
				 GENPOS_CART_REF *refId,
				 GENPOS_CART_CONFIG *posId, 
				 double period,
				 double *pdLongit, /* Ecart longitudinal */
				 double *pdTransv,  /* Ecart transversal */
				 double *pdTheta,   /* Ecart angulaire */
				 double *pSdLongit, /* Somme ecarts longit. */
				 double *pSdTransv, /* Somme ecarts transv. */
				 double *pSdTheta) /* Somme ecarts ang. */

{
  double dxA, dyA;              /* Erreurs position par rapport rep atelier */
  double dLongit;               /* Erreur de deplacement longitudinal */
  double dTransv;               /* Erreur de deplacement transversal */
  double dTheta=0;                /* Erreur de deplacement angulaire */
  static double SdLongit = 0.;   /* Somme des erreurs longitudinales */
  static double SdTransv = 0.;   /* Somme des erreurs transverses */
  static double SdTheta = 0.;      /* Somme des erreurs angulaires */
  double c, s;

  /* Donne'es positionnelles sur lesquelles on s'asservit */
  c = cos(posId->theta);
  s = sin(posId->theta);

  /* Obtenir les erreurs de position par rapport au rep atelier */
  /* point de controle */
  dxA = refId->x - posId->x;
  dyA = refId->y - posId->y;
  
  /* angles */
  *pdTheta = angleLimit(refId->theta - posId->theta);

  /* Obtenir les erreurs de deplacement transversal et longitudinal */
  *pdLongit = dLongit =   dxA * c + dyA * s;
  *pdTransv = dTransv = - dxA * s + dyA * c;

  /* Actualiser l'integrale des erreurs longitudinales */
  if (fabs (dLongit) < cmdId->longitErrorMax) {

    /* Actualiser la somme des erreurs */
    SdLongit += dLongit*period;
    
    /* La limiter */
    if (fabs (SdLongit) > cmdId->longitMaxErrSum)
      *pSdLongit = SdLongit = cmdId->longitMaxErrSum;
    else 
      *pSdLongit = SdLongit;
  }
  else 
    *pSdLongit = SdLongit = 0.;

  /* Actualiser l'integrale des erreurs transversales */
  if (fabs (dTransv) < cmdId->transvErrorMax) {
    
    /* Actualiser la somme des erreurs */
    SdTransv += dTransv*period;
    
    /* La limiter */
    if (fabs (SdTransv) > cmdId->transvMaxErrSum)
      *pSdTransv = SdTransv = cmdId->transvMaxErrSum;
    else
      *pSdTransv = SdTransv;
  }
  else 
    *pSdTransv = SdTransv = 0.;

  /* Actualiser l'integrale des erreurs angulaires */
  if (fabs (dTheta) < cmdId->angErrorMax) {
    
    /* Actualiser la somme des erreurs */
    SdTheta += dTheta*period;
    
    /* La limiter */
    if (fabs (SdTheta) > cmdId->angMaxErrSum)
      *pSdTheta = SdTheta = cmdId->angMaxErrSum;
    else
      *pSdTheta = SdTheta;
  }
  else 
    *pSdTheta = SdTheta = 0.;
}

/****************************************************************************
*
*   virtualRobotPositionGet -  
* 
*  This function computes the position of the virtual robot for a cart
*  towing a trailer. The trailer is hitched behind the wheel
*  axis of the cart and the virtual robot is the trailer. 
*
*  Return: OK ou ERROR
*
*/

static void 
virtualRobotPositionGetOffAxle(GENPOS_TRAILER_CONNECTION *kinParamPt,
			       GENPOS_TRCART_REF *refRobotId,
			       GENPOS_TRCART_CONFIG * posId,
			       GENPOS_CART_STATE * virtualRobotId,
			       GENPOS_CART_REF * virtualRefId)
{
  double x,y,theta,phi,x2,y2,theta2,l1, l2;
  double xRef,yRef,thetaRef,phiRef,x2Ref,y2Ref,theta2Ref;
  double vRef,wRef;

  l1 = kinParamPt->l1;
  l2 = kinParamPt->l2;  

  /* position du systeme */
  x = posId->robot.x;
  y = posId->robot.y;
  theta = posId->robot.theta;
  phi = posId->phi;
  /*
  v = robotId->v;
  w = robotId->w;
  */
  /* consigne du robot reel */
  xRef = refRobotId->robot.x;
  yRef = refRobotId->robot.y;
  thetaRef = refRobotId->robot.theta;
  phiRef = refRobotId->phi;
  vRef = refRobotId->robot.v;
  wRef = refRobotId->robot.w;

  /*Position du robot virtuel */
  theta2 = angleLimit(PI+theta+phi);
  x2 = x - l1*cos(theta) + l2*cos(theta2);
  y2 = y - l1*sin(theta) + l2*sin(theta2);

  /* Consigne du robot virtuel */
  /* OFF_AXLE */
  theta2Ref = angleLimit(PI+thetaRef+phiRef);
  x2Ref = xRef - l1*cos(thetaRef) + l2*cos(theta2Ref);
  y2Ref = yRef - l1*sin(thetaRef) + l2*sin(theta2Ref);

  /* mise a jour des donnees dans la SDI */
  /* position du robot virtuel */
  virtualRobotId->xRob = x2;
  virtualRobotId->yRob = y2;
  virtualRobotId->xRef = x2;
  virtualRobotId->yRef = y2;
  virtualRobotId->theta = theta2;
  /* OFF_AXLE */
  /*
  virtualRobotId->v = -v*cos(phi) + l1*w*sin(phi);
  virtualRobotId->w = -v*sin(phi)/l2 - l1*cos(phi)*w/l2;
  */

  /* consigne du robot virtuel */
  virtualRefId->x = x2Ref;
  virtualRefId->y = y2Ref;
  virtualRefId->theta = theta2Ref;
  /* OFF_AXLE */
  virtualRefId->v = -vRef*cos(phiRef) + l1*wRef*sin(phiRef);
  virtualRefId->w = -vRef*sin(phiRef)/l2 - l1*cos(phiRef)*wRef/l2;
  virtualRefId->backFlag = 1 - refRobotId->robot.backFlag;
}

/*
 * virtualRobotPositionGetCenteredAxle --
 *
 * This function computes the position of the virtual robot for a cart
 * towing a trailer. The trailer is hitched above the wheel axis of
 * the cart, the virtual robot is the symmetric of the cart wrt the
 * wheel axis of the trailer.
 *
 */

static void 
virtualRobotPositionGetCenteredAxle(GENPOS_TRAILER_CONNECTION *kinParamPt,
				    GENPOS_TRCART_REF *refRobotId,
				    GENPOS_TRCART_CONFIG * posId,
				    GENPOS_CART_STATE * virtualRobotId,
				    GENPOS_CART_REF * virtualRefId)
{
  double x,y,theta,phi,x2,y2,theta2,l1, l2;
  double xRef,yRef,thetaRef,phiRef,x2Ref,y2Ref,theta2Ref;
  double vRef,wRef;

  l1 = kinParamPt->l1;
  l2 = kinParamPt->l2;  

  /* position du systeme */
  x = posId->robot.x;
  y = posId->robot.y;
  theta = posId->robot.theta;
  phi = posId->phi;

  /*
  v = robotId->v;
  w = robotId->w;
  */
  /* consigne du robot reel */
  xRef = refRobotId->robot.x;
  yRef = refRobotId->robot.y;
  thetaRef = refRobotId->robot.theta;
  phiRef = refRobotId->phi;
  vRef = refRobotId->robot.v;
  wRef = refRobotId->robot.w;

  /*Position du robot virtuel */
  /* CENTERED_AXLE */
  x2 = x-2*l2*cos(theta+phi);
  y2 = y-2*l2*sin(theta+phi);
  theta2 = angleLimit(PI+theta+2*phi);

  /* Consigne du robot virtuel */
  /* CENTERED_AXLE */
  x2Ref = xRef-2*l2*cos(thetaRef+phiRef);
  y2Ref = yRef-2*l2*sin(thetaRef+phiRef);
  theta2Ref = angleLimit(PI+thetaRef+2*phiRef);

  /* mise a jour des donnees dans la SDI */
  /* position du robot virtuel */
  virtualRobotId->xRob = x2;
  virtualRobotId->yRob = y2;
  virtualRobotId->xRef = x2;
  virtualRobotId->yRef = y2;
  virtualRobotId->theta = theta2;
  /* CENTERED_AXLE */
  /*
  virtualRobotId->v = -v;
  virtualRobotId->w = -w-2*v/l2*sin(phi);
  */
  /* consigne du robot virtuel */
  virtualRefId->x = x2Ref;
  virtualRefId->y = y2Ref;
  virtualRefId->theta = theta2Ref;
  /* CENTERED_AXLE */
  virtualRefId->v = -vRef;
  virtualRefId->w = -wRef-2*vRef/l2*sin(phiRef);
}
